from typing import Generic, TypeVar

from rainbowadn.data.collection.collection_interface.collectioninterface import CollectionInterface
from rainbowadn.hashing.hashpoint import HashPoint
from rainbowadn.hashing.hashresolver import HashResolver
from rainbowadn.hashing.nullability.notnull import NotNull
from rainbowadn.hashing.nullability.null import Null
from rainbowadn.hashing.nullability.nullablereference import NullableReference
from rainbowadn.hashing.rainbow_factory import RainbowFactory

__all__ = ('BlockCollectionInterface',)

BlockType = TypeVar('BlockType')
HeaderType = TypeVar('HeaderType')
ActualStateType = TypeVar('ActualStateType')


class BlockCollectionInterface(
    CollectionInterface[
        BlockType
    ],
    Generic[BlockType, HeaderType, ActualStateType]
):
    def __init__(
            self,
            reference: NullableReference[BlockType],
            resolver: HashResolver
    ):
        assert isinstance(reference, NullableReference)
        assert isinstance(resolver, HashResolver)
        self.resolver = resolver
        super().__init__(reference)

    def _verify_link(
            self,
            block: BlockType,
            previous: NullableReference[BlockType]
    ) -> bool:
        raise NotImplementedError

    def verify_link(
            self,
            previous: NullableReference[BlockType]
    ) -> bool:
        assert isinstance(previous, NullableReference)
        if isinstance(self.reference.reference, Null):
            return True
        elif isinstance(self.reference.reference, NotNull):
            return self._verify_link(
                self.resolver.resolve(self.reference.reference.value),
                previous
            )
        else:
            raise TypeError

    def _base_state_factory(self) -> RainbowFactory[ActualStateType]:
        raise NotImplementedError

    def _base_state(self, block: BlockType) -> HashPoint[ActualStateType]:
        raise NotImplementedError

    def base_state(self) -> NullableReference[ActualStateType]:
        if isinstance(self.reference.reference, Null):
            return NullableReference(Null(), self._base_state_factory())
        elif isinstance(self.reference.reference, NotNull):
            return NullableReference.of(self._base_state(self.resolver.resolve(self.reference.reference.value)))
        else:
            raise TypeError

    def add(
            self,
            header: HashPoint[HeaderType]
    ) -> 'BlockCollectionInterface[BlockType, HeaderType, ActualStateType]':
        raise NotImplementedError
