from typing import Generic, TypeVar

from rainbowadn.chain.block import Block
from rainbowadn.chain.chaincollectionfactory import ChainCollectionFactory
from rainbowadn.chain.reduction.reduction import Reduction
from rainbowadn.chain.reduction.reductionchain import ReductionChain
from rainbowadn.chain.reduction.reductionprotocol import ReductionProtocol
from rainbowadn.chain.stages.stage import StateStage
from rainbowadn.hashing.nullability.nullablereference import NullableReference
from rainbowadn.hashing.rainbow_factory import RainbowFactory

__all__ = ('ReductionChainFactory',)

ReductorType = TypeVar('ReductorType')
AccumulatorType = TypeVar('AccumulatorType')


class ReductionChainFactory(
    ChainCollectionFactory[
        Block[
            ReductorType,
            StateStage[
                ReductorType,
                AccumulatorType,
                Reduction[ReductorType, AccumulatorType]
            ]
        ],
        ReductorType,
        AccumulatorType
    ],
    Generic[ReductorType, AccumulatorType]
):
    def __init__(
            self,
            reductor_factory: RainbowFactory[ReductorType],
            accumulator_factory: RainbowFactory[AccumulatorType],
            protocol: ReductionProtocol[ReductorType, AccumulatorType],
    ):
        assert isinstance(reductor_factory, RainbowFactory)
        assert isinstance(accumulator_factory, RainbowFactory)
        assert isinstance(protocol, ReductionProtocol)
        self.reductor_factory = reductor_factory
        self.accumulator_factory = accumulator_factory
        self.protocol = protocol

    def empty(self) -> ReductionChain[ReductorType, AccumulatorType]:
        return ReductionChain.empty(
            self.reductor_factory,
            self.accumulator_factory,
            self.protocol,
        )

    def from_reference(
            self, reference: NullableReference[
                Block[
                    ReductorType,
                    StateStage[
                        ReductorType,
                        AccumulatorType,
                        Reduction[ReductorType, AccumulatorType]
                    ]
                ],
            ]
            ) -> ReductionChain[
        ReductorType, AccumulatorType
    ]:
        assert isinstance(reference, NullableReference)
        return ReductionChain(
            reference,
            self.reductor_factory,
            self.accumulator_factory,
            self.protocol,
        )
