from typing import Generic, TypeVar

from rainbowadn.data.collection.collection_interface.activecollectioninterface import ActiveCollectionInterface
from rainbowadn.data.collection.stack.stack import Stack
from rainbowadn.hashing.hashpoint import HashPoint
from rainbowadn.hashing.hashresolver import HashResolver
from rainbowadn.hashing.nullability.notnull import NotNull
from rainbowadn.hashing.nullability.null import Null
from rainbowadn.hashing.nullability.nullablereference import NullableReference
from rainbowadn.hashing.rainbow_factory import RainbowFactory

__all__ = ('ActiveStack',)

ElementType = TypeVar('ElementType')


class ActiveStack(ActiveCollectionInterface[Stack[ElementType], ElementType], Generic[ElementType]):
    def __init__(self, reference: NullableReference[Stack[ElementType]]):
        assert isinstance(reference, NullableReference)
        super().__init__(reference)

    def add(self, element: HashPoint[ElementType]) -> 'ActiveStack[ElementType]':
        assert isinstance(element, HashPoint)
        return ActiveStack(NullableReference.off(Stack(self.reference, element)))

    @classmethod
    def empty(cls, factory: RainbowFactory[ElementType]) -> 'ActiveStack[ElementType]':
        assert isinstance(factory, RainbowFactory)
        return cls(NullableReference(Null(), Stack.factory(factory)))

    def str(self, resolver: HashResolver):
        assert isinstance(resolver, HashResolver)
        if isinstance(self.reference.reference, Null):
            return f'-'
        elif isinstance(self.reference.reference, NotNull):
            stack: Stack[ElementType] = resolver.resolve(self.reference.reference.value)
            assert isinstance(stack, Stack)
            return f'{ActiveStack(stack.previous).str(resolver)} {resolver.resolve(stack.element)}'
        else:
            raise TypeError
