from typing import Generic, TypeVar

from rainbowadn.data.collection.trees.binary.activebinarytree import ActiveBinaryTree
from rainbowadn.data.collection.trees.binary.binarytree import BinaryTree, BinaryTreeFactory
from rainbowadn.data.collection.trees.comparison.comparator import Comparator
from rainbowadn.data.collection.keymetadata import KeyMetadata, KeyMetadataFactory
from rainbowadn.hashing.hashpoint import HashPoint
from rainbowadn.hashing.nullability.null import Null
from rainbowadn.hashing.nullability.nullablereference import NullableReference
from rainbowadn.hashing.rainbow_factory import RainbowFactory

__all__ = ('UnbalancedABT',)

ActiveKeyType = TypeVar('ActiveKeyType')
MetaDataType = TypeVar('MetaDataType')


class UnbalancedABT(ActiveBinaryTree[ActiveKeyType, MetaDataType], Generic[ActiveKeyType, MetaDataType]):
    def __init__(
            self,
            comparator: Comparator[ActiveKeyType],
            reference: NullableReference[BinaryTree[KeyMetadata[ActiveKeyType, MetaDataType]]],
            metadata: HashPoint[MetaDataType]
    ):
        assert isinstance(comparator, Comparator)
        assert isinstance(reference, NullableReference)
        assert isinstance(metadata, HashPoint)
        super().__init__(comparator, reference)
        self._metadata = metadata

    def create(
            self, reference: NullableReference[BinaryTree[KeyMetadata[ActiveKeyType, MetaDataType]]]
    ) -> 'UnbalancedABT[ActiveKeyType, MetaDataType]':
        assert isinstance(reference, NullableReference)
        return type(self)(
            self.comparator,
            reference,
            self._metadata
        )

    @classmethod
    def empty(
            cls,
            comparator: Comparator[ActiveKeyType],
            metadata: HashPoint[MetaDataType],
            factory: RainbowFactory[ActiveKeyType]
    ):
        assert isinstance(comparator, Comparator)
        assert isinstance(metadata, HashPoint)
        assert isinstance(factory, RainbowFactory)
        return cls(
            comparator,
            NullableReference(
                Null(),
                BinaryTreeFactory(KeyMetadataFactory(factory, metadata.factory))
            ),
            metadata
        )

    def empty_metadata(self) -> HashPoint[MetaDataType]:
        return self._metadata

    def metadata(
            self,
            treel: NullableReference[BinaryTree[KeyMetadata[ActiveKeyType, MetaDataType]]],
            treer: NullableReference[BinaryTree[KeyMetadata[ActiveKeyType, MetaDataType]]],
            key: HashPoint[ActiveKeyType]
    ) -> HashPoint[MetaDataType]:
        assert isinstance(treel, NullableReference)
        assert isinstance(treer, NullableReference)
        assert isinstance(key, HashPoint)
        return self._metadata

    def balance(self) -> 'UnbalancedABT[ActiveKeyType, MetaDataType]':
        return self
