from typing import Generic, TypeVar

from rainbowadn.data.collection.trees.comparison.comparator import Comparator, Comparison
from rainbowadn.data.collection.keyed import Keyed
from rainbowadn.hashing.hashpoint import HashPoint
from rainbowadn.hashing.hashresolver import HashResolver

__all__ = ('KeyedComparator',)

ComparatorKeyType = TypeVar('ComparatorKeyType')


class KeyedComparator(
    Comparator[Keyed[ComparatorKeyType]], Generic[ComparatorKeyType]
):
    def __init__(self, resolver: HashResolver, comparator: Comparator[ComparatorKeyType]):
        assert isinstance(resolver, HashResolver)
        assert isinstance(comparator, Comparator)
        self.comparator = comparator
        super().__init__(resolver)

    def compare(
            self,
            original: HashPoint[Keyed[ComparatorKeyType]],
            key: HashPoint[Keyed[ComparatorKeyType]]
    ) -> Comparison:
        assert isinstance(original, HashPoint)
        assert isinstance(key, HashPoint)
        return self.comparator.compare(
            self.resolver.resolve(original).key,
            self.resolver.resolve(key).key,
        )
